/* Server Select (versione 2)*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#define DIM_BUFF 100
#define DIM_DIR 80
#define FILENAME_LENGTH 80
#define max(a,b) ((a) > (b) ? (a) : (b))

/********************************************************/

typedef struct
{
  char nomeFile[FILENAME_LENGTH];
  char car;
}
Request;

/********************************************************/

int conta_file (char *name)
{
  DIR *dir;
  struct dirent * dd;
  int count = 0;
  dir = opendir (name);
  while ((dd = readdir(dir)) != NULL)
  {
    printf("Trovato il file %s\n", dd-> d_name);
    count++;
  }
  printf("Numero totale di file %d\n", count);
  closedir (dir);
  return count;
}


/********************************************************/

void gestore(int signo)
{
  int stato;
  printf("esecuzione gestore di SIGCHLD\n");
  wait(&stato);
}


/********************************************************/

int main(int argc, char **argv)
{
  int  listenfd, connfd, udpfd, nready, maxfdp1;
  const int on = 1;
  char dir[DIM_DIR], ccorr;
  // Lunghezza del nome del file, piu' fine linea e terminatore
  // stringa
  char risposta[FILENAME_LENGTH+2];
  int ris;
  fd_set rset;
  struct hostent *host;
  struct hostent *clienthost;  
  int len, fd, port, nread, cont;
  struct sockaddr_in clientaddr, servaddr;
  Request req;
  DIR *direttorio;
  struct dirent * dd;



  /* CONTROLLO ARGOMENTI ---------------------------------- */
  if(argc!=2)
  {
    printf("Error: %s port\n", argv[0]);
    exit(1);
  }
  else port = atoi(argv[1]);

  printf("Server avviato\n");

  /* CREAZIONE SOCKET TCP ------------------------------------------------------ */
  listenfd=socket(AF_INET, SOCK_STREAM, 0);
  if (listenfd <0)
  {perror("apertura socket TCP "); exit(1);}
  printf("Creata la socket TCP d'ascolto, fd=%d\n", listenfd);

  /* INIZIALIZZAZIONE INDIRIZZO SERVER ----------------------------------------- */
  memset ((char *)&servaddr, 0, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = INADDR_ANY;
  servaddr.sin_port = htons(port);

  if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0)
  {perror("set opzioni socket TCP"); exit(2);}
  printf("Set opzioni socket TCP ok\n");

  if (bind(listenfd,(struct sockaddr *) &servaddr, sizeof(servaddr))<0)
  {perror("bind socket TCP"); exit(3);}
  printf("Bind socket TCP ok\n");

  if (listen(listenfd, 5)<0)
  {perror("listen"); exit(4);}
  printf("Listen ok\n");



  /* CREAZIONE SOCKET UDP ----------------------------------------------------- */
  udpfd=socket(AF_INET, SOCK_DGRAM, 0);
  if(udpfd <0)
  {perror("apertura socket UDP"); exit(5);}
  printf("Creata la socket UDP, fd=%d\n", udpfd);

  /* INIZIALIZZAZIONE INDIRIZZO SERVER ---------------------------------------- */
  memset ((char *)&servaddr, 0, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = INADDR_ANY;
  servaddr.sin_port = htons(port);

  if(setsockopt(udpfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0)
  {perror("set opzioni socket UDP"); exit(6);}
  printf("Set opzioni socket UDP ok\n");

  if(bind(udpfd,(struct sockaddr *) &servaddr, sizeof(servaddr))<0)
  {perror("bind socket UDP"); exit(7);}
  printf("Bind socket UDP ok\n");

  /* AGGANCIO GESTORE PER EVITARE FIGLI ZOMBIE -------------------------------- */
  signal(SIGCHLD, gestore);

  /* PULIZIA E SETTAGGIO MASCHERA DEI FILE DESCRIPTOR ------------------------- */
  FD_ZERO(&rset);
  maxfdp1=max(listenfd, udpfd)+1;


  /* CICLO DI RICEZIONE EVENTI DALLA SELECT ----------------------------------- */
  for(;;)
  {
    FD_SET(listenfd, &rset);
    FD_SET(udpfd, &rset);

    if ((nready=select(maxfdp1, &rset, NULL, NULL, NULL))<0)
    {
      if (errno==EINTR) continue;
      else {perror("select"); exit(8);}
    }


    /* GESTIONE RICHIESTE DI LS  ----------------------------- */

    if (FD_ISSET(listenfd, &rset))
    {
      printf("Ricevuta richiesta di ls\n");

      len = sizeof(struct sockaddr_in);
      if((connfd = accept(listenfd,(struct sockaddr *)&clientaddr,&len))<0)
      {
        if (errno==EINTR) continue;
        else {perror("accept"); exit(9);}
      }

      if (fork()==0)
      { /* processo figlio che serve la richiesta di operazione */
	    close(listenfd);
	    host=gethostbyaddr( (char *) &clientaddr.sin_addr, sizeof(clientaddr.sin_addr), AF_INET);
	    if (host == NULL){ printf("client host information not found\n"); }
	    else printf("Server (figlio): host client e' %s \n", host->h_name);

		// Leggo la richiesta del client
		if( read(connfd, &ccorr, 1)>0 )
		{
		  printf("Server (figlio): carattere richiesto: %c\n", ccorr);
		  if( read(connfd, dir, sizeof(dir))>0 )
		  {
		    // Apertura directory
            if( (direttorio = opendir (dir)) != NULL )
            {
              while ((dd = readdir(direttorio)) != NULL)
              {
                printf("Letto nome file: %s\n", dd->d_name);
                for(cont=0; cont<strlen(dd->d_name); cont++)
                {
                  if(dd->d_name[cont]==ccorr)
                  {
                  	printf("Invio %s che contiene il carattere richiesto\n", dd->d_name);
                  	strcpy(risposta, dd->d_name);
                  	strcat(risposta, "\n");
                  	printf("Risposta inviata: %s\n", risposta);
                    if (write(connfd, risposta, strlen(risposta))<0)
                    {perror("write"); break;}
                    else break;
                  }	
                }
              }
              // Liberiamo le risorse occupate
              closedir(direttorio);
              close(connfd);
            }
		  }
          else{ perror("read: "); }		  
		}
		else{ perror("read: "); }
        
      }//figlio

      /* padre chiude la socket dell'operazione */
      close(connfd);

    } /* fine gestione richieste di ls */



    /* GESTIONE RICHIESTE DI CARATTERE NELLA STRINGA -------------------- */

    if (FD_ISSET(udpfd, &rset))
    {
      len=sizeof(struct sockaddr_in);
      if (recvfrom(udpfd, &req, sizeof(Request), 0, (struct sockaddr *)&clientaddr, &len)<0)
      {perror("recvfrom "); continue;}	

      printf("Operazione richiesta, nome file: %s, carattere: %c\n", req.nomeFile, req.car);
      clienthost=gethostbyaddr( (char *) &clientaddr.sin_addr, sizeof(clientaddr.sin_addr), AF_INET);
      if (clienthost == NULL) printf("client host information not found\n");
	  else printf("Operazione richiesta da: %s %i\n", clienthost->h_name,
        (unsigned)ntohs(clientaddr.sin_port)); 

      /* Verifico l'esistenza del file */
      if((fd=open(req.nomeFile, O_RDONLY))<0)
      {
        perror("open file sorgente");
        ris=-1;
      }
      else{
        ris=0;
        while((nread=read(fd, &ccorr, 1))>0)
        {
          if(ccorr==req.car) ris++;
        }
      }
    
    
      // DEBUG
      printf("Risposta: %d\n", ris);
      ris=htonl(ris);

      if (sendto(udpfd, &ris, sizeof(ris), 0, (struct sockaddr *)&clientaddr, len)<0)
      {perror("sendto "); continue;}

    } /* fine gestione richieste di carattere */

  } /* ciclo for della select */

  /* NEVER ARRIVES HERE */
  exit(0);

}

